"""Classes for translating into the syntax of different markup languages"""

from __future__ import annotations

import sys
from collections.abc import Mapping
from contextlib import contextmanager
from typing import Any, Sequence

from skops.card._model_card import TableSection

if sys.version_info.minor >= 9:
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


class PandocItem(TypedDict):
    t: str
    c: dict


class Markdown:
    """Mapping of pandoc parsed document to Markdown

    This class has a ``mapping`` attribute, which is just a dict. The keys are
    Pandoc types and the values are functions that transform the corresponding
    value into a string with markdown syntax. Those functions are all prefixed
    with ``_``, e.g. ``_image`` for transforming a pandoc ``Image`` into a
    markdown figure, or ``_raw_block``, to transform a pandoc ``RawBlock``.

    From the caller side, only the ``__call__`` method should be used, the rest
    should be considered internals.

    """

    def __init__(self):
        # markdown syntax dispatch table
        self.mapping = {
            "Space": self._space,
            "Plain": self._plain,
            "Str": self._str,
            "Strong": self._strong,
            "Emph": self._emph,
            "Strikeout": self._strikeout,
            "RawInline": self._raw_inline,
            "RawBlock": self._raw_block,
            "SoftBreak": self._soft_break,
            "LineBreak": self._line_break,
            "Para": self._para,
            "Header": self._header,
            "Image": self._image,
            "Figure": self._figure,
            "CodeBlock": self._code_block,
            "Code": self._code,
            "Table": self._table,
            "Div": self._parse_div,
            "Link": self._link,
            "BulletList": self._bullet_list,
            "OrderedList": self._ordered_list,
            "Quoted": self._quoted,
            "BlockQuote": self._block_quote,
        }
        # Start indentation level at -1 because we want the first incremented
        # indentation level to be at 0. Otherwise we would need to keep track if
        # it's the first time and then don't increment, which is more
        # complicated.
        self._indent_trace = []

    @contextmanager
    def _indented(self, *, spaces: int):
        """Temporarily increment indentation by one"""
        self._indent_trace.append(spaces)
        yield
        self._indent_trace.pop(-1)

    def _get_indent(self, *, incr: int = 0) -> str:
        """Get current indentation, optionally incremented"""
        # TODO: explain why skipping 1st item
        return " " * (incr + sum(self._indent_trace[:-1]))

    @staticmethod
    def _space(value) -> str:
        return " "

    def _plain(self, value) -> str:
        parts = [self.__call__(subitem) for subitem in value]
        return "".join(parts)

    @staticmethod
    def _str(value) -> str:
        # escape \
        return value.replace("\\", "\\\\")

    def _strong(self, value) -> str:
        parts = ["**"]
        parts += [self.__call__(subitem) for subitem in value]
        parts.append("**")
        return "".join(parts)

    def _emph(self, value) -> str:
        parts = ["_"]
        parts += [self.__call__(subitem) for subitem in value]
        parts.append("_")
        return "".join(parts)

    def _strikeout(self, value) -> str:
        parts = ["~~"]
        parts += [self.__call__(subitem) for subitem in value]
        parts.append("~~")
        return "".join(parts)

    @staticmethod
    def _raw_inline(value) -> str:
        _, line = value
        return line

    def _raw_block(self, item) -> str:
        # throw away the first item, which is just something like 'html'
        # might have to revisit this if output != markdown
        _, line = item
        return line

    def _soft_break(self, value) -> str:
        incr = 0 if not self._indent_trace else self._indent_trace[-1]
        return "\n" + self._get_indent(incr=incr)

    def _line_break(self, value) -> str:
        return "\n"

    def _make_content(self, content):
        parts = []
        for item in content:
            part = "".join(self.__call__(item))
            parts.append(part)
        return "".join(parts)

    def _para(self, value: list[dict[str, str]]) -> str:
        content = self._make_content(value)
        return content

    def _header(self, value: tuple[int, Any, list[dict[str, str]]]) -> str:
        level, _, content_parts = value
        section_name = self._make_content(content_parts)
        return section_name

    def _image(self, value) -> str:
        (ident, _, keyvals), caption, (dest, typef) = value
        # it seems like ident and keyvals are not relevant for markdown

        if not caption:  # pragma: no cover
            # not sure if this can be reached, just to be safe
            raise ValueError("Figure missing a caption")

        if not typef.startswith("fig:"):  # pragma: no cover
            # not sure if this can be reached, just to be safe
            raise ValueError(f"Cannot deal with figure of type '{typef}'")

        caption = "".join(self.__call__(i) for i in caption)
        content = f"![{caption}]({dest})"
        return content

    def _figure(self, value) -> str:  # pragma: no cover
        # Figure type was added in Pandoc v3.0
        (ident, classes, keyvals), caption, (body,) = value

        body_type = body["t"]
        # we can only deal with plain figures for now
        if body_type != "Plain":
            raise ValueError(f"Cannot deal with figure of type '{body_type}'")

        plain_fig = body["c"][0]["c"]
        plain_fig[2][1] = "fig:"
        return self._image(plain_fig)

    @staticmethod
    def _code_block(item: tuple[tuple[int, list[str], list[str]], str]) -> str:
        # a codeblock consists of: (id, classes, namevals) contents
        (_, classes, _), content = item
        block_start = "```"
        if classes:
            block_start += ", ".join(classes)
        block_end = "```"
        content = "\n".join((block_start, content, block_end))
        return content

    @staticmethod
    def _code(item: tuple[Any, str]) -> str:
        _, txt = item
        return f"`{txt}`"

    def _table_cols_old(self, items) -> list[str]:
        columns = []
        for (content,) in items:
            column = self.__call__(content)
            columns.append(column)
        return columns

    def _table_cols_new(self, items) -> list[str]:  # pragma: no cover
        columns = []
        fn = self.__call__
        for item in items:
            _, alignment, _, _, content = item
            column = "".join(fn(part) for part in content)
            columns.append(column)
        return columns

    def _table_body_old(self, items) -> list[list[str]]:
        body = []
        for row_items in items:
            row = []
            for col_row_item in row_items:
                if not col_row_item:
                    content = ""
                else:
                    content = col_row_item[0]
                row.append(self.__call__(content))
            body.append(row)
        return body

    def _table_body_new(self, items) -> list[list[str]]:  # pragma: no cover
        body = []
        fn = self.__call__
        for _, row_items in items:
            row = []
            for col_row_item in row_items:
                _, alignment, _, _, content = col_row_item
                row.append("".join(fn(part) for part in content))
            body.append(row)
        return body

    def _table_old(self, item) -> tuple[list[str], list[list[str]]]:
        # pandoc < 2.10
        _, _, _, thead, tbody = item
        columns = self._table_cols_old(thead)
        body = self._table_body_old(tbody)
        return columns, body

    def _table_new(self, item) -> tuple[list[str], list[list[str]]]:  # pragma: no cover
        # pandoc >= 2.10
        # attr capt specs thead tbody tfoot
        _, _, _, thead, tbody, _ = item
        # header
        (_, thead_bodies) = thead
        (attr, thead_body) = thead_bodies[0]  # multiple headers?
        columns = self._table_cols_new(thead_body)
        # rows
        # attr rhc hd bd
        _, _, _, trows = tbody[0]  # multiple groups of rows?
        body = self._table_body_new(trows)
        return columns, body

    def _table(self, item) -> str:
        if len(item) == 6:  # pragma: no cover
            # pandoc >= 2.5
            columns, body = self._table_new(item)
        else:
            # pandoc < 2.5
            columns, body = self._table_old(item)

        table: Mapping[str, Sequence[Any]]
        if not body:
            table = {key: [] for key in columns}
        else:
            # body is row oriented, transpose to column oriented
            data_transposed = zip(*body)
            table = {key: val for key, val in zip(columns, data_transposed)}

        res = TableSection(title="", content="", table=table).format()
        return res

    def _parse_div(self, item) -> str:
        # note that in markdown, we basically just use the raw html
        (ident, classes, kvs), contents = item

        # build div tag
        tags = ["<div"]
        if ident:
            tags.append(f' id="{ident}"')
        if classes:
            classes = " ".join(classes)
            tags.append(f' class="{classes}"')
        if kvs:
            kvparts = []
            for k, v in kvs:
                if not v:  # e.g. just ['hidden', '']
                    kvparts.append(k)
                else:
                    kvparts.append(f'{k}="{v}"')
            tags.append(f' {" ".join(kvparts)}')
        tags.append(">")

        start = "".join(tags)
        middle = []
        for content in contents:
            with self._indented(spaces=2):
                middle.append(self.__call__(content))
        end = "</div>"
        return "".join([start] + middle + [end])

    def _link(self, item) -> str:
        _, txt, (src, _) = item
        txt_formatted = self._make_content(txt)
        return f"[{txt_formatted}]({src})"

    def _make_list_item(self, items: str, list_marker: str):
        # helper function used for bullet and ordered lists
        parts = [self.__call__(subitem) for subitem in items]
        content = "\n".join(parts)
        return f"{self._get_indent()}{list_marker} {content}"

    def _bullet_list(self, item) -> str:
        # we don't differentiate between lists starting with "-", "*", or "+".
        list_marker = "-"
        parts = []
        # bullet lists use 2 spaces for indentation to align "- "
        with self._indented(spaces=2):
            for subitem in item:
                parts.append(self._make_list_item(subitem, list_marker=list_marker))
        return "\n".join(parts)

    def _ordered_list(self, item) -> str:
        # we don't make use of num_type and sep_type, which just indicates that
        # numbers are presented as decimal numbers using a period
        (start, num_type, sep_type), items = item
        parts = []
        # ordered lists use 3 spaces for indentation to align "1. "
        with self._indented(spaces=3):
            for i, subitem in enumerate(items, start=start):
                parts.append(self._make_list_item(subitem, list_marker=f"{i}."))
        return "\n".join(parts)

    def _quoted(self, item: tuple[dict[str, str], list[PandocItem]]) -> str:
        quote_type, content = item
        type_ = quote_type["t"]
        try:
            sym = {"DoubleQuote": '"', "SingleQuote": "'"}[type_]
        except KeyError as exc:  # pragma: no cover
            # can probably not be reached, but let's be sure
            msg = (
                f"The parsed document contains '{type_}', which is not "
                "supported yet, please open an issue on GitHub"
            )
            raise ValueError(msg) from exc

        text = "".join(self.__call__(i) for i in content)
        return f"{sym}{text}{sym}"

    def _block_quote(self, item: list[PandocItem]) -> str:
        parts = []
        for subitem in item:
            content = self.__call__(subitem)
            # add quote symbolx
            content = content.replace("\n", "\n> ")
            parts.append(content)

        # add a quote symbol to the very start
        text = "> " + "\n> ".join(parts)
        return text

    def __call__(self, item: str | PandocItem) -> str:
        if isinstance(item, str):
            return item

        type_, value = item["t"], item.get("c")
        try:
            res = self.mapping[type_](value)
        except KeyError as exc:
            msg = (
                f"The parsed document contains '{type_}', which is not "
                "supported yet, please open an issue on GitHub"
            )
            raise ValueError(msg) from exc

        # recursively call until the value has been resolved into a str
        return self.__call__(res)



from ._hf_hub import (
    add_files,
    download,
    get_config,
    get_model_output,
    get_requirements,
    init,
    push,
    update_env,
)

__all__ = [
    "add_files",
    "download",
    "get_config",
    "get_requirements",
    "get_model_output",
    "init",
    "push",
    "update_env",
]



"""All minimum dependencies for scikit-learn."""
import argparse

PYTEST_MIN_VERSION = "5.0.1"

# 'build' and 'install' is included to have structured metadata for CI.
# It will NOT be included in setup's extras_require
# The values are (version_spec, comma separated tags, condition)
# tags can be: 'build', 'install', 'docs', 'examples', 'tests', 'benchmark',
# 'rich'
# example:
#     "tomli": ("1.1.0", "install", "python_full_version < '3.11.0a7'"),
dependent_packages = {
    "scikit-learn": ("0.24", "install", None),
    "scikit-learn-intelex": ("2021.7.1", "docs", None),
    "huggingface_hub": ("0.10.1", "install", None),
    "tabulate": ("0.8.8", "install", None),
    "pytest": (PYTEST_MIN_VERSION, "tests", None),
    "pytest-cov": ("2.9.0", "tests", None),
    "flake8": ("3.8.2", "tests", None),
    "types-requests": ("2.28.5", "tests", None),
    "flaky": ("3.7.0", "tests", None),
    "sphinx": ("3.2.0", "docs", None),
    "sphinx-gallery": ("0.7.0", "docs", None),
    "sphinx-rtd-theme": ("1", "docs", None),
    "numpydoc": ("1.0.0", "docs", None),
    "sphinx-prompt": ("1.3.0", "docs", None),
    "sphinx-issues": ("1.2.0", "docs", None),
    "matplotlib": ("3.3", "docs, tests", None),
    "packaging": ("17.0", "install", None),
    "pandas": ("1", "docs, tests", None),
    # required for persistence tests of external libraries
    "lightgbm": ("3", "tests", None),
    "xgboost": ("1.6", "tests", None),
    # TODO: remove condition when catboost supports python 3.11
    "catboost": ("1.0", "tests", "python_version < '3.11'"),
    "fairlearn": ("0.7.0", "docs, tests", None),
    "rich": ("12", "tests, rich", None),
}


# create inverse mapping for setuptools
tag_to_packages: dict = {
    extra: []
    for extra in ["build", "install", "docs", "examples", "tests", "benchmark", "rich"]
}
for package, (min_version, extras, condition) in dependent_packages.items():
    for extra in extras.split(", "):
        spec = f"{package}>={min_version}"
        if condition:
            spec += f"; {condition}"
        tag_to_packages[extra].append(spec)


# Used by CI to get the min dependencies
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Get min dependencies for a package")

    parser.add_argument("package", choices=dependent_packages)
    args = parser.parse_args()
    min_version = dependent_packages[args.package][0]
    print(min_version)



import numpy as np
import scipy
from sklearn.utils import all_estimators

from ._utils import get_type_name

PRIMITIVES_TYPES = [int, float, str, bool]

PRIMITIVE_TYPE_NAMES = ["builtins." + t.__name__ for t in PRIMITIVES_TYPES]

SKLEARN_ESTIMATOR_TYPE_NAMES = [
    get_type_name(estimator_class)
    for _, estimator_class in all_estimators()
    if get_type_name(estimator_class).startswith("sklearn.")
]

SCIPY_UFUNC_TYPE_NAMES = sorted(
    set(
        [
            get_type_name(getattr(scipy.special, attr))
            for attr in dir(scipy.special)
            if isinstance(getattr(scipy.special, attr), np.ufunc)
            and get_type_name(getattr(scipy.special, attr)).startswith("scipy")
        ]
    )
)



"""The current protocol of the skops version

Notes on updating the protocol:

Every time that a backwards incompatible change to the skops format is made
for the first time within a release, the protocol should be bumped to the next
higher number. The old version of the Node, which knows how to deal with the
old state, should be preserved, registered, and tested. Let's give an example:

- There is a BC breaking change in FunctionNode.
- Since it's the first BC breaking change in the skops format in this release,
  bump skops.io._protocol.PROTOCOL (this file) from version X to X+1.
- Move the old FunctionNode code into 'skops/io/old/_general_vX.py', where 'X'
  is the old protocol.
- Register the _general_vX.FunctionNode in NODE_TYPE_MAPPING inside of
  _persist.py.
- Write a test in test_persist_old.py that shows that the old state can
  still be loaded. Look at test_persist_old.test_function_v0 for inspiration.

Now, if a user loads a FunctionNode state with version X using skops with
version Y>X, the old code will be used instead of the new one. For all other
node types, if there is no loader for version X, skops will automatically use
version Y instead.

"""
PROTOCOL = 1



from __future__ import annotations

import argparse
import logging
import os
import pathlib
import pickle
from typing import Optional

from skops.cli._utils import get_log_level
from skops.io import dumps, get_untrusted_types


def _convert_file(
    input_file: os.PathLike,
    output_file: os.PathLike,
    logger: logging.Logger = logging.getLogger(),
) -> None:
    """Function that is called by ``skops convert`` entrypoint.

    Loads a pickle model from the input path, converts to skops format, and saves to
    output file.

    Parameters
    ----------
    input_file : os.PathLike
        Path of input .pkl model to load.

    output_file : os.PathLike
        Path to save .skops model to.

    """
    model_name = pathlib.Path(input_file).stem

    logger.debug(f"Converting {model_name}")

    with open(input_file, "rb") as f:
        obj = pickle.load(f)
    skops_dump = dumps(obj)

    untrusted_types = get_untrusted_types(data=skops_dump)

    if not untrusted_types:
        logger.info(f"No unknown types found in {model_name}.")
    else:
        untrusted_str = ", ".join(untrusted_types)

        logger.warning(
            f"While converting {input_file}, "
            "the following unknown types were found: "
            f"{untrusted_str}. "
            f"When loading {output_file} with skops.load, these types must be "
            "specified as 'trusted'"
        )

    with open(output_file, "wb") as out_file:
        logger.debug(f"Writing to {output_file}")
        out_file.write(skops_dump)


def format_parser(
    parser: Optional[argparse.ArgumentParser] = None,
) -> argparse.ArgumentParser:
    """Adds arguments and help to parent CLI parser for the convert method."""

    if not parser:  # used in tests
        parser = argparse.ArgumentParser()

    parser_subgroup = parser.add_argument_group("convert")
    parser_subgroup.add_argument("input", help="Path to an input file to convert. ")

    parser_subgroup.add_argument(
        "-o",
        "--output-file",
        help=(
            "Specify the output file name for the converted skops file. "
            "If not provided, will default to using the same name as the input file, "
            "and saving to the current working directory with the suffix '.skops'."
        ),
        default=None,
    )
    parser_subgroup.add_argument(
        "-v",
        "--verbose",
        help=(
            "Increases verbosity of logging. Can be used multiple times to increase "
            "verbosity further."
        ),
        action="count",
        dest="loglevel",
        default=0,
    )
    return parser


def main(
    parsed_args: argparse.Namespace,
) -> None:
    output_file = parsed_args.output_file
    input_file = parsed_args.input

    logging.basicConfig(
        format="%(levelname)-8s: %(message)s", level=get_log_level(parsed_args.loglevel)
    )

    if not output_file:
        # No filename provided, defaulting to base file path
        file_name = pathlib.Path(input_file).stem
        output_file = pathlib.Path.cwd() / f"{file_name}.skops"

    _convert_file(
        input_file=input_file,
        output_file=output_file,
    )



"""Tests for skops.io.visualize"""

from unittest.mock import Mock, patch

import pytest
import sklearn
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import FeatureUnion, Pipeline
from sklearn.preprocessing import (
    FunctionTransformer,
    MinMaxScaler,
    PolynomialFeatures,
    StandardScaler,
)

import skops.io as sio


class TestVisualizeTree:
    @pytest.fixture
    def simple(self):
        return MinMaxScaler(feature_range=(-555, 123))

    @pytest.fixture
    def pipeline(self):
        def unsafe_function(x):
            return x

        # fmt: off
        pipeline = Pipeline([
            ("features", FeatureUnion([
                ("scaler", StandardScaler()),
                ("scaled-poly", Pipeline([
                    ("polys", FeatureUnion([
                        ("poly1", PolynomialFeatures()),
                        ("poly2", PolynomialFeatures(degree=3, include_bias=False))
                    ])),
                    ("square-root", FunctionTransformer(unsafe_function)),
                    ("scale", MinMaxScaler()),
                ])),
            ])),
            ("clf", LogisticRegression(random_state=0, solver="liblinear")),
        ]).fit([[0, 1], [2, 3], [4, 5]], [0, 1, 2])
        # fmt: on
        return pipeline

    @pytest.mark.parametrize("show", ["all", "trusted", "untrusted"])
    def test_print_simple(self, simple, show, capsys):
        file = sio.dumps(simple)
        sio.visualize(file, show=show)

        # Output is the same for "all" and "trusted" because all nodes are
        # trusted. Colors are not recorded by capsys.
        expected = [
            "root: sklearn.preprocessing._data.MinMaxScaler",
            "└── attrs: builtins.dict",
            "    ├── feature_range: builtins.tuple",
            "    │   ├── content: json-type(-555)",
            "    │   └── content: json-type(123)",
            "    ├── copy: json-type(true)",
            "    ├── clip: json-type(false)",
            '    └── _sklearn_version: json-type("{}")'.format(sklearn.__version__),
        ]
        if show == "untrusted":
            # since no untrusted, only show root
            expected = expected[:1]

        stdout, _ = capsys.readouterr()
        assert stdout.strip() == "\n".join(expected)

    def test_print_pipeline(self, pipeline, capsys):
        file = sio.dumps(pipeline)
        sio.visualize(file)

        # no point in checking the whole output with > 120 lines
        expected_start = [
            "root: sklearn.pipeline.Pipeline",
            "└── attrs: builtins.dict",
            "    ├── steps: builtins.list",
            "    │   ├── content: builtins.tuple",
            '    │   │   ├── content: json-type("features")',
        ]
        expected_end = [
            "    ├── memory: json-type(null)",
            "    ├── verbose: json-type(false)",
            '    └── _sklearn_version: json-type("{}")'.format(sklearn.__version__),
        ]

        stdout, _ = capsys.readouterr()
        assert stdout.startswith("\n".join(expected_start))
        assert stdout.rstrip().endswith("\n".join(expected_end))

    def test_unsafe_nodes(self, pipeline):
        file = sio.dumps(pipeline)
        nodes = []

        def sink(nodes_iter, *args, **kwargs):
            nodes.extend(nodes_iter)

        sio.visualize(file, sink=sink)
        nodes_self_unsafe = [node for node in nodes if not node.is_self_safe]
        nodes_unsafe = [node for node in nodes if not node.is_safe]

        # there are currently 2 unsafe nodes, a numpy int and the custom
        # functions. The former might be considered safe in the future, in which
        # case this test needs to be changed.
        assert len(nodes_self_unsafe) == 2
        assert nodes_self_unsafe[0].val == "numpy.int64"
        assert nodes_self_unsafe[1].val == "test_visualize.unsafe_function"

        # it's not easy to test the number of indirectly unsafe nodes, because
        # it will depend on the nesting structure; we can only be sure that it's
        # more than 2, and one of them should be the FunctionTransformer
        assert len(nodes_unsafe) > 2
        assert any("FunctionTransformer" in node.val for node in nodes_unsafe)

    @pytest.mark.parametrize(
        "kwargs",
        [
            {},
            {"use_colors": False},
            {"tag_unsafe": "<careful>", "color_unsafe": "blue"},
        ],
    )
    def test_custom_print_config_passed_to_sink(self, simple, kwargs):
        # check that arguments are passed to sink
        def my_sink(nodes_iter, show, **sink_kwargs):
            for key, val in kwargs.items():
                assert sink_kwargs[key] == val

        file = sio.dumps(simple)
        sio.visualize(file, sink=my_sink, **kwargs)

    def test_custom_tags(self, simple, capsys):
        class UnsafeType:
            pass

        simple.copy = UnsafeType

        file = sio.dumps(simple)
        sio.visualize(file, tag_safe="NICE", tag_unsafe="OHNO")
        expected = [
            "root: sklearn.preprocessing._data.MinMaxScaler NICE",
            "└── attrs: builtins.dict NICE",
            "    ├── feature_range: builtins.tuple NICE",
            "    │   ├── content: json-type(-555) NICE",
            "    │   └── content: json-type(123) NICE",
            "    ├── copy: test_visualize.UnsafeType OHNO",
            "    ├── clip: json-type(false) NICE",
            '    └── _sklearn_version: json-type("{}") NICE'.format(
                sklearn.__version__
            ),
        ]

        stdout, _ = capsys.readouterr()
        assert stdout.strip() == "\n".join(expected)

    def test_custom_colors(self, simple):
        # test that custom colors are used in node representation, requires rich
        # to work
        pytest.importorskip("rich")

        class UnsafeType:
            pass

        simple.copy = UnsafeType
        file = sio.dumps(simple)

        # Colors are not recorded by capsys, so we cannot use it and must mock
        # printing
        mock_print = Mock()
        with patch("rich.print", mock_print):
            sio.visualize(
                file,
                color_safe="black",
                color_unsafe="cyan",
                color_child_unsafe="orange3",
            )

        mock_print.assert_called()

        calls = mock_print.call_args_list
        # The root node is indirectly unsafe through child
        assert (
            calls[0].args[0]
            == "root: [orange3]sklearn.preprocessing._data.MinMaxScaler"
        )
        # 'feature_range' is safe
        assert calls[6].args[0] == " feature_range: [black]builtins.tuple"
        # 'copy' is unsafe
        assert calls[15].args[0] == " copy: [cyan]test_visualize.UnsafeType [UNSAFE]"

    @pytest.mark.usefixtures("rich_not_installed")
    def test_no_colors_if_rich_not_installed(self, simple):
        # this test is similar to the previous one, except that we test that the
        # colors are *not* used if rich is not installed
        file = sio.dumps(simple)

        # don't use capsys, because it wouldn't capture the colors, thus need to
        # use mock
        mock_print = Mock()
        with patch("builtins.print", mock_print):
            sio.visualize(
                file,
                color_safe="black",
                color_unsafe="cyan",
                color_child_unsafe="orange3",
            )
        mock_print.assert_called()

        # check that none of the colors are being used
        colors = ("black", "cyan", "orange3")
        for call in mock_print.call_args_list:
            for color in colors:
                assert color not in call.args[0]

    def test_no_colors_if_use_colors_false(self, simple):
        # this test is similar to the previous one, except that we test that the
        # colors are *not* used, even if rich is installed, when passing
        # use_colors=False
        file = sio.dumps(simple)

        # don't use capsys, because it wouldn't capture the colors, thus need to
        # use mock
        mock_print = Mock()
        with patch("rich.print", mock_print):
            sio.visualize(
                file,
                color_safe="black",
                color_unsafe="cyan",
                color_child_unsafe="orange3",
                use_colors=False,
            )
        mock_print.assert_called()

        # check that none of the colors are being used
        colors = ("black", "cyan", "orange3")
        for call in mock_print.call_args_list:
            for color in colors:
                assert color not in call.args[0]

    def test_from_file(self, simple, tmp_path, capsys):
        f_name = tmp_path / "estimator.skops"
        sio.dump(simple, f_name)
        sio.visualize(f_name)

        expected = [
            "root: sklearn.preprocessing._data.MinMaxScaler",
            "└── attrs: builtins.dict",
            "    ├── feature_range: builtins.tuple",
            "    │   ├── content: json-type(-555)",
            "    │   └── content: json-type(123)",
            "    ├── copy: json-type(true)",
            "    ├── clip: json-type(false)",
            '    └── _sklearn_version: json-type("{}")'.format(sklearn.__version__),
        ]
        stdout, _ = capsys.readouterr()
        assert stdout.strip() == "\n".join(expected)



import json
import os
import pickle
import re
import shutil
import tempfile
import warnings
from importlib import metadata
from pathlib import Path
from uuid import uuid4

import numpy as np
import pandas as pd
import pytest
import sklearn
from flaky import flaky
from huggingface_hub import HfApi
from huggingface_hub.utils import RepositoryNotFoundError
from sklearn.datasets import load_diabetes, load_iris
from sklearn.linear_model import LinearRegression, LogisticRegression

from skops import card
from skops.hub_utils import (
    add_files,
    download,
    get_config,
    get_model_output,
    get_requirements,
    init,
    push,
    update_env,
)
from skops.hub_utils._hf_hub import (
    _create_config,
    _get_column_names,
    _get_example_input_from_tabular_data,
    _get_example_input_from_text_data,
    _validate_folder,
)
from skops.hub_utils.tests.common import HF_HUB_TOKEN
from skops.io import dump

iris = load_iris(as_frame=True, return_X_y=False)
diabetes = load_diabetes(as_frame=True, return_X_y=False)

IS_SKLEARN_DEV_BUILD = "dev" in sklearn.__version__


@pytest.fixture
def temp_path():
    with tempfile.TemporaryDirectory(prefix="skops-test-temp-path") as temp_path:
        yield temp_path


@pytest.fixture(scope="session")
def repo_path():
    with tempfile.TemporaryDirectory(prefix="skops-test-sample-repo") as repo_path:
        yield Path(repo_path)


@pytest.fixture
def destination_path():
    with tempfile.TemporaryDirectory(prefix="skops-test") as dir_path:
        yield Path(dir_path)


def get_classifier():
    X, y = iris.data, iris.target
    clf = LogisticRegression(solver="newton-cg").fit(X, y)
    return clf


def get_regressor():
    X, y = diabetes.data, diabetes.target
    model = LinearRegression().fit(X, y)
    return model


@pytest.fixture(scope="session")
def classifier(repo_path, config_json):
    # Create a simple model file for the purpose of testing
    clf = get_classifier()
    config_path, file_format = config_json
    model_file = CONFIG[file_format]["sklearn"]["model"]["file"]
    path = repo_path / model_file

    try:
        if file_format == "pickle":
            with open(path, "wb") as f:
                pickle.dump(clf, f)
        elif file_format == "skops":
            dump(clf, path)
        yield path
    finally:
        path.unlink(missing_ok=True)


CONFIG = {
    "pickle": {
        "sklearn": {
            "environment": ['scikit-learn="1.1.1"'],
            "model": {"file": "model.pickle"},
        }
    },
    "skops": {
        "sklearn": {
            "environment": ['scikit-learn="1.1.1"'],
            "model": {"file": "model.skops"},
        }
    },
}


@pytest.fixture(scope="session", params=["skops", "pickle"])
def config_json(repo_path, request):
    path = repo_path / "config.json"
    try:
        with open(path, "w") as f:
            json.dump(CONFIG[request.param], f)
        yield path, request.param
    finally:
        path.unlink(missing_ok=True)


def test_validate_format(classifier):
    dir_path = tempfile.mkdtemp()
    shutil.rmtree(dir_path)
    with pytest.raises(ValueError, match="Cannot determine the input file*"):
        init(
            model=classifier,
            requirements=["scikit-learn"],
            dst=dir_path,
            task="tabular-classification",
            data=iris.data,
            model_format="dummy",
        )


def test_validate_folder(config_json):
    config_path, file_format = config_json
    _, file_path = tempfile.mkstemp()
    dir_path = tempfile.mkdtemp()
    with pytest.raises(TypeError, match="The given path is not a directory."):
        _validate_folder(path=file_path)

    with pytest.raises(TypeError, match="Configuration file `config.json` missing."):
        _validate_folder(path=dir_path)

    with open(Path(dir_path) / "config.json", "w") as f:
        json.dump(dict(), f)

    with pytest.raises(
        TypeError, match="Model file not configured in the configuration file."
    ):
        _validate_folder(path=dir_path)

    shutil.copy2(config_path, dir_path)
    model_file = CONFIG[file_format]["sklearn"]["model"]["file"]
    with pytest.raises(TypeError, match=f"Model file {model_file} does not exist."):
        _validate_folder(path=dir_path)

    (Path(dir_path) / model_file).touch()

    # this should now work w/o an error
    _validate_folder(path=dir_path)


@pytest.mark.parametrize(
    "data, task, expected_config",
    [
        (
            iris.data,
            "tabular-classification",
            {
                "sklearn": {
                    "columns": [
                        "petal length (cm)",
                        "petal width (cm)",
                        "sepal length (cm)",
                        "sepal width (cm)",
                    ],
                    "environment": ['scikit-learn="1.1.1"', "numpy"],
                    "example_input": {
                        "petal length (cm)": [1.4, 1.4, 1.3],
                        "petal width (cm)": [0.2, 0.2, 0.2],
                        "sepal length (cm)": [5.1, 4.9, 4.7],
                        "sepal width (cm)": [3.5, 3.0, 3.2],
                    },
                    "model": {"file": "model.pkl"},
                    "task": "tabular-classification",
                }
            },
        ),
        (
            ["test", "text", "problem", "random"],
            "text-classification",
            {
                "sklearn": {
                    "environment": ['scikit-learn="1.1.1"', "numpy"],
                    "example_input": {"data": ["test", "text", "problem"]},
                    "model": {"file": "model.pkl"},
                    "task": "text-classification",
                }
            },
        ),
    ],
)
def test_create_config(data, task, expected_config):
    dir_path = tempfile.mkdtemp()
    _create_config(
        model_path="model.pkl",
        requirements=['scikit-learn="1.1.1"', "numpy"],
        dst=dir_path,
        task=task,
        data=data,
    )

    with open(Path(dir_path) / "config.json") as f:
        config = json.load(f)
        for key in ["environment", "model", "task"]:
            assert config["sklearn"][key] == expected_config["sklearn"][key]

        keys = ["example_input"]
        if "tabular" in task:
            # text data doesn't introduce any "columns" in the configuration
            keys += ["columns"]
        for key in keys:
            assert sorted(config["sklearn"][key]) == sorted(
                expected_config["sklearn"][key]
            )


def test_create_config_invalid_text_data(temp_path):
    with pytest.raises(ValueError, match="The data needs to be a sequence of strings."):
        _create_config(
            model_path="model.pkl",
            requirements=['scikit-learn="1.1.1"', "numpy"],
            task="text-classification",
            data=[1, 2, 3],
            dst=temp_path,
        )


def test_atomic_init(classifier, temp_path):
    with pytest.raises(ValueError):
        # this fails since we're passing an invalid task.
        init(
            model=classifier,
            requirements=["scikit-learn"],
            dst=temp_path,
            task="tabular-classification",
            data="invalid",
        )

    # this passes even though the above init has failed once, on the same
    # destination path.
    init(
        model=classifier,
        requirements=["scikit-learn"],
        dst=temp_path,
        task="tabular-classification",
        data=iris.data,
    )


def test_init_invalid_task(classifier, temp_path):
    with pytest.raises(
        ValueError, match="Task invalid not supported. Supported tasks are"
    ):
        init(
            model=classifier,
            requirements=["scikit-learn"],
            dst=temp_path,
            task="invalid",
            data=iris.data,
        )


def test_init(classifier, config_json):
    config_path, file_format = config_json
    # create a temp directory and delete it, we just need a unique name.
    dir_path = tempfile.mkdtemp()
    shutil.rmtree(dir_path)

    version = metadata.version("scikit-learn")
    init(
        model=classifier,
        requirements=[f'scikit-learn="{version}"'],
        dst=dir_path,
        task="tabular-classification",
        data=iris.data,
    )
    _validate_folder(path=dir_path)

    # it should fail a second time since the folder is no longer empty.
    with pytest.raises(OSError, match="None-empty dst path already exists!"):
        init(
            model=classifier,
            requirements=[f'scikit-learn="{version}"'],
            dst=dir_path,
            task="tabular-classification",
            data=iris.data,
        )


def test_init_no_warning_or_error(classifier, config_json):
    config_path, file_format = config_json
    # for the happy path, there should be no warning
    dir_path = tempfile.mkdtemp()
    shutil.rmtree(dir_path)
    version = metadata.version("scikit-learn")

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        init(
            model=classifier,
            requirements=[f'scikit-learn="{version}"'],
            dst=dir_path,
            task="tabular-classification",
            data=iris.data,
        )


def test_model_file_does_not_exist_raises(repo_path, config_json):
    config_path, file_format = config_json
    # when the model file does not exist, raise an OSError
    model_path = repo_path / "foobar.pickle"
    dir_path = tempfile.mkdtemp()
    shutil.rmtree(dir_path)
    version = metadata.version("scikit-learn")

    msg = re.escape(f"Model file '{model_path}' does not exist.")
    with pytest.raises(OSError, match=msg):
        init(
            model=model_path,
            requirements=[f'scikit-learn="{version}"'],
            dst=dir_path,
            task="tabular-classification",
            data=iris.data,
        )
    model_path.unlink(missing_ok=True)


def test_init_empty_model_file_errors(repo_path, config_json):
    config_path, file_format = config_json
    # when model file is empty, warn users
    model_path = Path(repo_path / "foobar.pickle")
    model_path.touch()

    dir_path = tempfile.mkdtemp()
    shutil.rmtree(dir_path)
    version = metadata.version("scikit-learn")

    with pytest.raises(
        RuntimeError, match=re.escape(f"Model file '{model_path}' is empty.")
    ):
        init(
            model=model_path,
            requirements=[f'scikit-learn="{version}"'],
            dst=dir_path,
            task="tabular-classification",
            data=iris.data,
        )
    model_path.unlink(missing_ok=True)


@pytest.mark.network
@flaky(max_runs=3)
@pytest.mark.parametrize("explicit_create", [True, False])
def test_push_download(
    explicit_create,
    repo_path,
    destination_path,
    classifier,
    config_json,
):
    config_path, file_format = config_json
    client = HfApi()

    version = metadata.version("scikit-learn")
    init(
        model=classifier,
        requirements=[f'scikit-learn="{version}"'],
        dst=destination_path,
        task="tabular-classification",
        data=iris.data,
    )

    user = client.whoami(token=HF_HUB_TOKEN)["name"]
    repo_id = f"{user}/test-{uuid4()}"
    if explicit_create:
        client.create_repo(repo_id=repo_id, token=HF_HUB_TOKEN, repo_type="model")
    push(
        repo_id=repo_id,
        source=repo_path,
        token=HF_HUB_TOKEN,
        commit_message="test message",
        create_remote=True,
        private=True,
    )

    # TODO: remove 1st message when huggingface_hub < v0.12 is dropped
    # message changes in huggingface_hub v0.12, test both
    match = (
        "If the repo is private, make sure you are authenticated"
        "|"
        "If you are trying to access a private or gated repo, "
        "make sure you are authenticated"
    )
    with pytest.raises(RepositoryNotFoundError, match=match):
        download(repo_id=repo_id, dst="/tmp/test")

    with pytest.raises(OSError, match="None-empty dst path already exists!"):
        download(repo_id=repo_id, dst=destination_path, token=HF_HUB_TOKEN)

    files = client.list_repo_files(repo_id=repo_id, use_auth_token=HF_HUB_TOKEN)
    for f_name in [classifier.name, config_path.name]:
        assert f_name in files

    try:
        with tempfile.TemporaryDirectory(prefix="skops-test") as dst:
            download(repo_id=repo_id, dst=dst, token=HF_HUB_TOKEN, keep_cache=False)
            copy_files = os.listdir(dst)
            assert set(copy_files) == set(files)
    finally:
        client.delete_repo(repo_id=repo_id, token=HF_HUB_TOKEN)


@pytest.fixture
def repo_path_for_inference():
    # Create a separate path for test_inference so that the test does not have
    # any side-effect on existing tests
    with tempfile.TemporaryDirectory(prefix="skops-test-sample-repo") as repo_path:
        yield Path(repo_path)


@pytest.mark.network
@pytest.mark.inference
@pytest.mark.skipif(
    IS_SKLEARN_DEV_BUILD, reason="Inference tests cannot run with sklearn dev build"
)
@flaky(max_runs=3)
@pytest.mark.parametrize(
    "model_func, data, task",
    [
        (get_classifier, iris, "tabular-classification"),
        (get_regressor, diabetes, "tabular-regression"),
    ],
    ids=["classifier", "regressor"],
)
def test_inference(
    config_json,
    model_func,
    data,
    task,
    repo_path_for_inference,
    destination_path,
):
    # test inference backend for classifier and regressor models. This test can
    # take a lot of time and be flaky.
    config_path, file_format = config_json
    if file_format != "pickle":
        pytest.skip(
            f"Inference only supports pickle at the moment. Given format: {file_format}"
        )

    client = HfApi()
    repo_path = repo_path_for_inference
    model_file = CONFIG[file_format]["sklearn"]["model"]["file"]
    model = model_func()
    model_path = repo_path / model_file

    with open(model_path, "wb") as f:
        pickle.dump(model, f)

    version = metadata.version("scikit-learn")
    init(
        model=model_path,
        requirements=[f'scikit-learn="{version}"'],
        dst=destination_path,
        task=task,
        data=data.data,
    )

    # TODO: remove when card init at repo init is merged
    model_card = card.Card(
        model, metadata=card.metadata_from_config(Path(destination_path))
    )
    model_card.save(Path(destination_path) / "README.md")

    user = client.whoami(token=HF_HUB_TOKEN)["name"]
    repo_id = f"{user}/test-{uuid4()}"

    push(
        repo_id=repo_id,
        source=destination_path,
        token=HF_HUB_TOKEN,
        commit_message="test message",
        create_remote=True,
        # api-inference doesn't support private repos for community projects.
        private=False,
    )

    X_test = data.data.head(5)
    y_pred = model.predict(X_test)
    output = get_model_output(repo_id, data=X_test, token=HF_HUB_TOKEN)

    # cleanup
    client.delete_repo(repo_id=repo_id, token=HF_HUB_TOKEN)
    model_path.unlink(missing_ok=True)

    assert np.allclose(output, y_pred)


def test_get_config(repo_path, config_json):
    config_path, file_format = config_json
    config = get_config(repo_path)

    assert config == CONFIG[file_format]
    assert get_requirements(repo_path) == ['scikit-learn="1.1.1"']


def test_update_env(repo_path, config_json):
    # sanity check
    assert get_requirements(repo_path) == ['scikit-learn="1.1.1"']
    update_env(path=repo_path, requirements=['scikit-learn="1.1.2"'])
    assert get_requirements(repo_path) == ['scikit-learn="1.1.2"']


def test_get_example_input_from_tabular_data():
    with pytest.raises(
        ValueError,
        match=(
            "The data is not a pandas.DataFrame, a 2D numpy.ndarray or a "
            "list/tuple that can be converted to a 2D numpy.ndarray."
        ),
    ):
        _get_example_input_from_tabular_data("random")
    with pytest.raises(ValueError):
        _get_example_input_from_tabular_data(["a", "b", "c"])

    examples = _get_example_input_from_tabular_data(np.ones((5, 10)))
    # the result is a dictionary of column name: list of values
    assert len(examples) == 10
    assert len(examples["x0"]) == 3

    examples = _get_example_input_from_tabular_data(np.ones((5, 10)).tolist())
    # the result is a dictionary of column name: list of values
    assert len(examples) == 10
    assert len(examples["x0"]) == 3

    examples = _get_example_input_from_tabular_data(
        pd.DataFrame(np.ones((5, 10)), columns=[f"column{x}" for x in range(10)])
    )
    # the result is a dictionary of column name: list of values
    assert len(examples) == 10
    assert len(examples["column0"]) == 3


@pytest.mark.parametrize(
    "data, expected_length",
    [
        (["a", "b", "c", "d"], 3),
        (np.array(["a", "b", "c", "d"]), 3),
        (set(["a", "b", "c", "d"]), 3),
        (tuple(["a", "b", "c", "d"]), 3),
        (["a"], 1),
        ([], 0),
    ],
)
def test_get_example_input_from_text_data(data, expected_length):
    example_input = _get_example_input_from_text_data(data)
    assert len(example_input["data"]) == expected_length


@pytest.mark.parametrize("data", ["random", [1, 2, 3], 420])
def test_get_example_input_from_text_data_invalid_text_data(data):
    with pytest.raises(ValueError, match="The data needs to be a sequence of strings."):
        _get_example_input_from_text_data(data)


def test_get_column_names():
    with pytest.raises(
        ValueError,
        match=(
            "The data is not a pandas.DataFrame, a 2D numpy.ndarray or a "
            "list/tuple that can be converted to a 2D numpy.ndarray."
        ),
    ):
        _get_column_names("random")
    with pytest.raises(ValueError):
        _get_column_names(["a", "b", "c"])

    X_array = np.ones((5, 10), dtype=np.float32)
    expected_columns = [f"x{x}" for x in range(10)]
    assert _get_column_names(X_array) == expected_columns

    expected_columns = [f"column{x}" for x in range(10)]
    X_df = pd.DataFrame(X_array, columns=expected_columns)
    assert _get_column_names(X_df) == expected_columns


def test_get_example_input_from_tabular_data_pandas_not_installed(pandas_not_installed):
    # use pandas_not_installed fixture from conftest.py to pretend that pandas
    # is not installed and check that the function does not raise when pandas
    # import fails
    _get_example_input_from_tabular_data(np.ones((5, 10)))


def test_get_column_names_pandas_not_installed(pandas_not_installed):
    # use pandas_not_installed fixture from conftest.py to pretend that pandas
    # is not installed and check that the function does not raise when pandas
    # import fails
    _get_column_names(np.ones((5, 10)))


class TestAddFiles:
    @pytest.fixture
    def init_path(self, classifier, config_json):
        # create temporary directory
        dir_path = tempfile.mkdtemp()
        shutil.rmtree(dir_path)

        version = metadata.version("scikit-learn")
        init(
            model=classifier,
            requirements=[f'scikit-learn="{version}"'],
            dst=dir_path,
            task="tabular-classification",
            data=iris.data,
        )
        yield dir_path

    @pytest.fixture
    def some_file_0(self, temp_path):
        filename = Path(temp_path) / "file0.txt"
        with open(filename, "w") as f:
            f.write("")
        yield filename

    @pytest.fixture
    def some_file_1(self, temp_path):
        filename = Path(temp_path) / "file1.txt"
        with open(filename, "w") as f:
            f.write("")
        yield filename

    def test_adding_one_file_path(self, init_path, some_file_0):
        add_files(some_file_0, dst=init_path)
        assert os.path.exists(Path(init_path) / some_file_0.name)

    def test_adding_two_file_paths(self, init_path, some_file_0, some_file_1):
        add_files(some_file_0, some_file_1, dst=init_path)
        assert os.path.exists(Path(init_path) / some_file_0.name)
        assert os.path.exists(Path(init_path) / some_file_1.name)

    def test_adding_one_file_str(self, init_path, some_file_0):
        add_files(str(some_file_0), dst=init_path)
        assert os.path.exists(Path(init_path) / some_file_0.name)

    def test_adding_two_files_str(self, init_path, some_file_0, some_file_1):
        add_files(str(some_file_0), str(some_file_1), dst=init_path)
        assert os.path.exists(Path(init_path) / some_file_0.name)
        assert os.path.exists(Path(init_path) / some_file_1.name)

    def test_adding_str_and_path(self, init_path, some_file_0, some_file_1):
        add_files(str(some_file_0), some_file_1, dst=init_path)
        assert os.path.exists(Path(init_path) / some_file_0.name)
        assert os.path.exists(Path(init_path) / some_file_1.name)

    def test_dst_does_not_exist_raises(self, some_file_0):
        dst = tempfile.mkdtemp()
        shutil.rmtree(dst)
        msg = (
            rf"Could not find \'{re.escape(dst)}\', did you run "
            r"\'skops.hub_utils.init\' first\?"
        )
        with pytest.raises(FileNotFoundError, match=msg):
            add_files(some_file_0, dst=dst)

    def test_file_does_not_exist_raises(self, init_path, some_file_0):
        non_existing_file = "foobar.baz"
        msg = r"File \'foobar.baz\' could not be found."
        with pytest.raises(FileNotFoundError, match=msg):
            add_files(some_file_0, non_existing_file, dst=init_path)

    def test_adding_existing_file_works_if_exist_ok(self, init_path, some_file_0):
        add_files(some_file_0, dst=init_path)
        assert os.path.exists(Path(init_path) / some_file_0.name)
        add_files(some_file_0, dst=init_path, exist_ok=True)
        assert os.path.exists(Path(init_path) / some_file_0.name)

    def test_adding_existing_file_raises(self, init_path, some_file_0):
        # first time around no warning
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            add_files(some_file_0, dst=init_path, exist_ok=False)

        msg = (
            f"File '{re.escape(some_file_0.name)}' already found "
            f"at '{re.escape(init_path)}'."
        )
        with pytest.raises(FileExistsError, match=msg):
            add_files(some_file_0, dst=init_path)


class TestUseIntelex:
    # Tests related to the usage of scikit-learn intelex, see #251
    def make_config(self, model, requirements, **kwargs):
        dir_path = tempfile.mkdtemp()
        shutil.rmtree(dir_path)

        init(
            model=model,
            dst=dir_path,
            task="tabular-classification",
            data=iris.data,
            requirements=requirements,
            **kwargs,
        )
        config = get_config(dir_path)
        return config

    def test_no_intelex(self, classifier):
        # by default, intelex is not being used
        config = self.make_config(model=classifier, requirements=["foobar"])
        environement = config["sklearn"]["environment"]

        assert config["sklearn"]["use_intelex"] is False
        assert not any(r.startswith("scikit-learn-intelex") for r in environement)

    def test_use_intelex_but_not_explicitly_in_requirements(self, classifier):
        # when using intelex, if it's not explicitly in the environment, add it
        # automatically
        config = self.make_config(
            model=classifier, requirements=["foobar"], use_intelex=True
        )
        environement = config["sklearn"]["environment"]

        assert config["sklearn"]["use_intelex"] is True
        assert any(r == "scikit-learn-intelex" for r in environement)

    def test_use_intelex_explicitly_in_requirements(self, classifier):
        # when users specify intelex explicitly, it's not added automatically to
        # the requirements
        reqs = ["foobar", "scikit-learn-intelex==2023.0.0"]
        config = self.make_config(model=classifier, requirements=reqs, use_intelex=True)
        environement = config["sklearn"]["environment"]

        assert config["sklearn"]["use_intelex"] is True
        assert not any(r == "scikit-learn-intelex" for r in environement)
        assert any(r == "scikit-learn-intelex==2023.0.0" for r in environement)



from __future__ import annotations

from typing import Any, Sequence

import numpy as np

from skops.io._audit import Node
from skops.io._utils import LoadContext, gettype

PROTOCOL = 0


class RandomGeneratorNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        self.children = {"bit_generator_state": state["content"]["bit_generator"]}
        self.trusted = self._get_trusted(trusted, [np.random.Generator])

    def _construct(self):
        # first restore the state of the bit generator
        bit_generator = gettype(
            "numpy.random", self.children["bit_generator_state"]["bit_generator"]
        )()
        bit_generator.state = self.children["bit_generator_state"]

        # next create the generator instance
        return gettype(self.module_name, self.class_name)(bit_generator=bit_generator)


NODE_TYPE_MAPPING = {
    ("RandomGeneratorNode", PROTOCOL): RandomGeneratorNode,
}



#! /usr/bin/env python
# License: 3-clause BSD
import builtins

from setuptools import setup

# This is a bit (!) hackish: we are setting a global variable so that the
# main modelcard __init__ can detect if it is being loaded by the setup
# routine, to avoid attempting to load components.
builtins.__SKOPS_SETUP__ = True  # type: ignore


import skops  # noqa
import skops._min_dependencies as min_deps  # noqa

VERSION = skops.__version__

DISTNAME = "skops"
DESCRIPTION = (
    "A set of tools to push scikit-learn based models to and pull from Hugging Face Hub"
)
with open("README.rst") as f:
    LONG_DESCRIPTION = f.read()
MAINTAINER = "Adrin Jalali"
MAINTAINER_EMAIL = "adrin.jalali@gmail.com"
URL = "http://github.com/skops-dev/skops"
DOWNLOAD_URL = "https://pypi.org/project/skops/#files"
LICENSE = "MIT"
PROJECT_URLS = {
    "Bug Tracker": "http://github.com/skops-dev/skops/issues",
    "Documentation": "http://github.com/skops-dev/skops",
    "Source Code": "http://github.com/skops-dev/skops",
}


def setup_package():
    package_data = dict(
        entry_points={
            "console_scripts": [
                "skops = skops.cli.entrypoint:main_cli",
            ],
        }
    )

    metadata = dict(
        name=DISTNAME,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        license=LICENSE,
        url=URL,
        download_url=DOWNLOAD_URL,
        project_urls=PROJECT_URLS,
        version=VERSION,
        long_description=LONG_DESCRIPTION,
        classifiers=[
            "Intended Audience :: Science/Research",
            "Intended Audience :: Developers",
            "License :: OSI Approved",
            "Programming Language :: Python",
            "Topic :: Software Development",
            "Topic :: Scientific/Engineering",
            "Development Status :: 1 - Planning",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: Unix",
            "Operating System :: MacOS",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: Implementation :: CPython",
        ],
        python_requires=">=3.8",
        install_requires=min_deps.tag_to_packages["install"],
        extras_require={
            "docs": min_deps.tag_to_packages["docs"],
            "tests": min_deps.tag_to_packages["tests"],
            "rich": min_deps.tag_to_packages["rich"],
        },
        include_package_data=True,
    )

    setup(**package_data, **metadata)


if __name__ == "__main__":
    setup_package()



import argparse

import skops.cli._convert


def main_cli(command_line_args=None):
    """Main command line interface entrypoint for all command line Skops methods.

    To add a new entrypoint:
        1. Create a new method to call that accepts a namespace
        2. Create a new subparser formatter to define the expected CL arguments
        3. Add those to the function map.
    """
    entry_parser = argparse.ArgumentParser(
        prog="Skops",
        description="Main entrypoint for all command line Skops methods.",
        add_help=True,
    )

    subparsers = entry_parser.add_subparsers(
        title="Commands",
        description="Skops command to call",
        dest="cmd",
        help="Sub-commands help",
    )

    # function_map should map a command to
    #   method: the command to call (gets set to default 'func')
    #   format_parser: the function used to create a subparser for that command
    function_map = {
        "convert": {
            "method": skops.cli._convert.main,
            "format_parser": skops.cli._convert.format_parser,
        },
    }

    for func_name, values in function_map.items():
        # Add subparser for each function in func map,
        # and assigns default func to be "method" from function_map
        subparser = subparsers.add_parser(func_name)
        subparser.set_defaults(func=values["method"])
        values["format_parser"](subparser)

    # Parse arguments with arg parser for given function in function map,
    # Then call the matching method in the function_map with the argument namespace
    args = entry_parser.parse_args(command_line_args)
    args.func(args)



"""Persistence tests for old versions of the protocol"""

from __future__ import annotations

import numpy as np
import pytest
from scipy import special
from sklearn.preprocessing import FunctionTransformer

from skops.io import dumps, loads
from skops.io._utils import get_module
from skops.io.tests._utils import (
    assert_method_outputs_equal,
    assert_params_equal,
    downgrade_state,
)

#############
# VERSION 0 #
#############


def dummy_func(X):
    return X


@pytest.mark.parametrize("func", [np.sqrt, len, special.exp10, dummy_func])
def test_persist_function_v0(func):
    call_count = 0

    # function_get_state as it was for protocol 0
    def old_function_get_state(obj, save_context):
        # added for testing
        nonlocal call_count
        call_count += 1
        # end

        res = {
            "__class__": obj.__class__.__name__,
            "__module__": get_module(obj),
            "__loader__": "FunctionNode",
            "content": {
                "module_path": get_module(obj),
                "function": obj.__name__,
            },
        }
        return res

    estimator = FunctionTransformer(func=func)
    X, y = [0, 1], [2, 3]
    estimator.fit(X, y)

    dumped = dumps(estimator)
    # importent: downgrade the state to mimic older version
    downgraded = downgrade_state(
        data=dumped,
        keys=["content", "content", "func"],
        old_state=old_function_get_state(func, None),
        protocol=0,
    )
    loaded = loads(downgraded, trusted=True)

    # sanity check: ensure that the old get_state function was really called
    assert call_count == 1

    # check that loaded estimator is identical
    assert_params_equal(estimator.__dict__, loaded.__dict__)
    assert_method_outputs_equal(estimator, loaded, X)


@pytest.mark.parametrize(
    "rng",
    [
        np.random.default_rng(),
        np.random.Generator(np.random.PCG64DXSM(seed=123)),
    ],
    ids=["default_rng", "Generator"],
)
def test_random_generator_v0(rng):
    call_count = 0

    # random_generator_get_state as it was for protocol 0
    def old_random_generator_get_state(obj, save_context):
        # added for testing
        nonlocal call_count
        call_count += 1
        # end

        bit_generator_state = obj.bit_generator.state
        res = {
            "__class__": obj.__class__.__name__,
            "__module__": get_module(type(obj)),
            "__loader__": "RandomGeneratorNode",
            "content": {"bit_generator": bit_generator_state},
        }
        return res

    rng.random(123)  # move RNG forwards
    dumped = dumps(rng)
    # importent: downgrade the whole state to mimic older version
    downgraded = downgrade_state(
        data=dumped,
        keys=None,
        old_state=old_random_generator_get_state(rng, None),
        protocol=0,
    )

    # old loader only worked with trusted=True, see #329
    loaded = loads(downgraded, trusted=True)

    # sanity check: ensure that the old get_state function was really called
    assert call_count == 1

    rand_floats_expected = rng.random(100)
    rand_floats_loaded = loaded.random(100)
    np.testing.assert_equal(rand_floats_loaded, rand_floats_expected)



from __future__ import annotations

from typing import Any, Sequence

from skops.io._audit import Node
from skops.io._trusted_types import SCIPY_UFUNC_TYPE_NAMES
from skops.io._utils import LoadContext, _import_obj

PROTOCOL = 0


class FunctionNode(Node):
    def __init__(
        self,
        state: dict[str, Any],
        load_context: LoadContext,
        trusted: bool | Sequence[str] = False,
    ) -> None:
        super().__init__(state, load_context, trusted)
        # TODO: what do we trust?
        self.trusted = self._get_trusted(trusted, default=SCIPY_UFUNC_TYPE_NAMES)
        self.children = {"content": state["content"]}

    def _construct(self):
        return _import_obj(
            self.children["content"]["module_path"],
            self.children["content"]["function"],
        )

    def _get_function_name(self) -> str:  # pragma: no cover
        return (
            self.children["content"]["module_path"]
            + "."
            + self.children["content"]["function"]
        )

    def get_unsafe_set(self) -> set[str]:  # pragma: no cover
        if (self.trusted is True) or (self._get_function_name() in self.trusted):
            return set()

        return {self._get_function_name()}


NODE_TYPE_MAPPING = {
    ("FunctionNode", PROTOCOL): FunctionNode,
}
